-- Broker control implementation
-- File name: broker_control.ads
-- Author: Jaime Prez Crespo <japecre@gsyc.escet.urjc.es>
-- Last modification: June 13, 2005
--
-- Copyright (c) 2005, Jaime Prez Crespo
-- All rights reserved.
-- 
-- Redistribution and use in source and binary forms, with or without modification,
-- are permitted provided that the following conditions are met:
-- - Redistributions of source code must retain the above copyright notice, this list
--   of conditions and the following disclaimer.
-- - Redistributions in binary form must reproduce the above copyright notice, this list
--   of conditions and the following disclaimer in the documentation and/or other materials
--   provided with the distribution.
-- - Neither the name of the Rey Juan Carlos University nor the names of its contributors
--   may be used to endorse or promote products derived from this software without specific
--   prior written permission.
-- 
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
-- OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
-- AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
-- CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
-- DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-- DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
-- IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-- OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

with ada.text_io;
with ada.numerics.discrete_random;
with ada.strings.unbounded;
with ada.exceptions;
with ada.characters.latin_1;
with ada.strings.fixed;
with lists_generic;
with broker_data;
with broker_messages;
with http_common;
with utils;
with handlers;

package body broker_control is

    package udp renames lower_layer_udp;
    package unb renames ada.strings.unbounded;
    package utl renames utils;
    package hdl renames handlers;
    use ada.strings.fixed;
    use broker_data;
    use broker_messages;

    function build_response(r: response) return string is
        str: unb.unbounded_string;
    begin
        case r is
        when ROOT =>
            str := http_common.status_line(http_common.OK,http_common.HTTP11);
            http_common.header_line(str,http_common.SERVER,broker_identification);
            http_common.header_line(str,http_common.CONTENTLEN,
                                    unb.to_unbounded_string(integer'image(html_login'length)));
            unb.append(str,ascii.cr & ascii.lf);
            unb.append(str,html_login);
        when REDIRECT =>
            str := http_common.status_line(http_common.MOVEDTEMP,http_common.HTTP11);
            http_common.header_line(str,http_common.SERVER,broker_identification);
            -- this is a special response. We need to specify a location header with
            -- variable data, so leave it to caller
        when OPTIONS =>
            str := http_common.status_line(http_common.OK,http_common.HTTP11);
            http_common.header_line(str,http_common.SERVER,broker_identification);
            -- this is a special response. We need to specify a content-length header with
            -- variable data, so leave it to caller
        when ERROR =>
            str := http_common.status_line(http_common.OK,http_common.HTTP11);
            http_common.header_line(str,http_common.SERVER,broker_identification);
            http_common.header_line(str,http_common.CONTENTLEN,
                                    unb.to_unbounded_string(integer'image(html_error'length)));
            unb.append(str,ascii.cr & ascii.lf);
            unb.append(str,html_error);
        end case;
        return unb.to_string(str);
    end;

    function http(request: http_request.request) return string is
        use type unb.unbounded_string;
        use type offer_kind;
        use type ol.reference;  -- for comparing references
        use type al.reference;
        str,r,t, color_left, color_right, hr: unb.unbounded_string;
        index: natural;
        sid: integer;
        s: auth.session;
        data: auth.auth_info_addr;
        omsg: message_offer;
        amsg: message_accept;
        buf: aliased udp.buffer_type(1024);
        oref: ol.reference := ol.null_reference;
        aref: al.reference := al.null_reference;
        offer: offer_queue;
        accep: accept_queue;
        hr_done: boolean := false;
    begin
        if unb.to_string(request.uri.path) = "/" then
            return build_response(ROOT);
        elsif unb.to_string(request.uri.path) = "/login" then
            -- initialize session
            s := auth.new_session;
            str := unb.to_unbounded_string(build_response(REDIRECT));
            http_common.header_line(str,http_common.LOCATION,
                                    unb.to_unbounded_string(integer'image(auth.get_id(s)) &
                                    ".html"));
            http_common.header_line(str,http_common.CONTENTLEN,
                                    unb.to_unbounded_string(integer'image(0)));
            unb.append(str,ascii.cr & ascii.lf);

            -- do authentication
            data := new user_data;
            data.login := unb.to_unbounded_string(http_common.get_post_var("name",request.data));

            auth.associate(s,data);
            if not auth.authenticate(s) then
                str := unb.to_unbounded_string(build_response(ROOT));
                utl.error("http: unauthorized user tried to login");
            end if;

            return unb.to_string(str);
        elsif unb.to_string(request.uri.path) = "/manage" then
            -- do authentication
            sid := integer'value(http_common.get_post_var("session",request.data));
            s := auth.validate(sid);
            data := auth.get_info(s);

            -- build message
            if http_common.get_post_var("type",request.data) = "compra" then
                omsg.kind := BUY;
            else
                omsg.kind := SELL;
            end if;
            omsg.quantity := integer'value(http_common.get_post_var("quantity",request.data));
            omsg.stock_price := integer'value(http_common.get_post_var("value",request.data));
            omsg.offer_id := integer'value(http_common.get_post_var("offer",request.data));
            omsg.offerer_id := data.login;
            omsg.round_id := broker_data.round;

            -- new round started, update accounting
            -- 1 per action per round
            while data.registered > 0 and data.registered < round loop
                data.money := data.money + data.stock;
                data.registered := data.registered + 1;
            end loop;

            -- check if user has enough room to perform the operation, and do it
            if (omsg.kind = BUY and then (omsg.quantity * omsg.stock_price <= data.money
                and omsg.quantity > 0 and omsg.stock_price > 0)) or 
               (omsg.kind = SELL and then (omsg.quantity <= data.vstock
                and omsg.stock_price > 0 and omsg.quantity > 0)) then
                udp.reset(buf);
                message'class'output(buf'access,message'class(omsg));
                udp.send(multicast_main_addr,buf'access);

                -- reflect changes on user data
                if omsg.kind = BUY then
                    data.money := data.money - omsg.quantity * omsg.stock_price;
                    data.vstock := data.vstock + omsg.quantity; 
                else
                    data.money := data.money + omsg.quantity * omsg.stock_price;
                    data.vstock := data.vstock - omsg.quantity;
                end if;
                data.offer := data.offer + 1;
                data.registered := round;
            -- elsewhere, do nothing!
            end if;

            str := unb.to_unbounded_string(build_response(REDIRECT));
            http_common.header_line(str,http_common.LOCATION,
                                    unb.to_unbounded_string(integer'image(auth.get_id(s)) &
                                    ".html"));
            http_common.header_line(str,http_common.CONTENTLEN,
                                    unb.to_unbounded_string(integer'image(0)));
            unb.append(str,ascii.cr & ascii.lf);

            return unb.to_string(str);
        elsif unb.to_string(request.uri.path) = "/accept" then
            -- do authentication
            sid := integer'value(http_common.get_post_var("session",request.data));
            s := auth.validate(sid);
            data := auth.get_info(s);

            -- new round started, update accounting
            -- 1 per action per round
            while data.registered > 0 and data.registered < round loop
                data.money := data.money + data.stock;
                data.registered := data.registered + 1;
            end loop;

            amsg.offer_id := integer'value(http_common.get_post_var("offer",request.data));
            amsg.offerer_id := unb.to_unbounded_string(http_common.get_post_var("offerer",request.data));
            amsg.acceptor_id := data.login;

            begin
                oref := search(amsg.offer_id,amsg.offerer_id);
                offer := ol.get_element(offers,oref);

                -- check if user has enough room to perform the operation, and do it
                if (offer.offer.kind = BUY and then offer.offer.quantity <= data.vstock) or
                   (offer.offer.kind = SELL and then
                    offer.offer.quantity * offer.offer.stock_price <= data.money) then
                    udp.reset(buf);
                    message'class'output(buf'access,message'class(amsg));
                    udp.send(multicast_main_addr,buf'access);

                    -- reflect changes on user data
                    if offer.offer.kind = BUY then
                        data.money := data.money + offer.offer.quantity * offer.offer.stock_price;
                        data.vstock := data.vstock - offer.offer.quantity; 
                    else
                        data.money := data.money - offer.offer.quantity * offer.offer.stock_price;
                        data.vstock := data.vstock + offer.offer.quantity;
                    end if;
                    data.registered := round;
                end if;
            exception
                when e: ebadoffer =>
                    utl.error("http: accept received for inexistent or old offer");
            end;

            str := unb.to_unbounded_string(build_response(REDIRECT));
            http_common.header_line(str,http_common.LOCATION,
                                    unb.to_unbounded_string(integer'image(auth.get_id(s)) &
                                    ".html"));
            http_common.header_line(str,http_common.CONTENTLEN,
                                    unb.to_unbounded_string(integer'image(0)));
            unb.append(str,ascii.cr & ascii.lf);

            return unb.to_string(str);
        else    -- any other page, maybe ok (session_id.html or xxx), or not
            str := request.uri.path;
            index := unb.index(str,".html");
            if index = 0 then
                -- invalid page!
                return build_response(ERROR);
            end if;
            unb.delete(str,index,index + 4);  -- .html
            unb.delete(str,1,1);      -- /

            s := auth.validate(integer'value(unb.to_string(str)));
            data := auth.get_info(s);

            -- new round started, update accounting
            -- 1 per action per round
            while data.registered > 0 and data.registered < round loop
                data.money := data.money + data.stock;
                data.registered := data.registered + 1;
            end loop;

            str := unb.to_unbounded_string(build_response(OPTIONS));
            -- build response body
            r := unb.to_unbounded_string(html_uopts);
            -- put real username
            index := unb.index(r,"$USERNAME$");
            unb.replace_slice(r,index,index - 1 +
                              unb.length(unb.to_unbounded_string("$USERNAME$")),
                              unb.to_string(data.login));
            -- put user stocks
            index := unb.index(r,"$STOCKS$");
            unb.replace_slice(r,index,index - 1 +
                              unb.length(unb.to_unbounded_string("$STOCKS$")),
                              trim(integer'image(data.stock),ada.strings.both));
            -- put user earned money
            index := unb.index(r,"$MONEY$");
            unb.replace_slice(r,index,index - 1 +
                              unb.length(unb.to_unbounded_string("$MONEY$")),
                              trim(integer'image(data.money),ada.strings.both));
            -- put the actual game round
            index := unb.index(r,"$ROUND$");
            unb.replace_slice(r,index,index - 1 +
                              unb.length(unb.to_unbounded_string("$ROUND$")),
                              trim(integer'image(round),ada.strings.both));
            -- put the offer identifier
            index := unb.index(r,"$OFFERID$");
            unb.replace_slice(r,index,index - 1 +
                              unb.length(unb.to_unbounded_string("$OFFERID$")),
                              trim(integer'image(data.offer),ada.strings.both));
            -- put the session identifier
            index := unb.index(r,"$SESSIONID$");
            unb.replace_slice(r,index,index - 1 +
                              unb.length(unb.to_unbounded_string("$SESSIONID$")),
                              trim(integer'image(auth.get_id(s)),ada.strings.both));

            -- build the offer lists
            if not ol.is_empty(offers) then
                oref := ol.last(offers);
            end if;

            while oref /= ol.null_reference loop
                -- get offer data
                offer := ol.get_element(offers,oref);
                if offer.offer.kind = BUY then
                    t := unb.to_unbounded_string("compra");
                else
                    t := unb.to_unbounded_string("venta");
                end if;

                -- check if we've done this offer, and put in red if so
                if offer.offer.offerer_id = data.login then
                    color_left := unb.to_unbounded_string("<font color=""red"">");
                    color_right := unb.to_unbounded_string("</font>");
                else
                    color_left := unb.to_unbounded_string("");
                    color_right := unb.to_unbounded_string("");
                end if;

                -- check if there is a round change in this iteration
                if offer.offer.round_id < round and not hr_done then
                    hr := unb.to_unbounded_string("<hr>Ronda anterior:<br>");
                    hr_done := true;
                else
                    hr := unb.to_unbounded_string("");
                end if;

                if offer.offer.round_id = round then
                    -- generate html
                    index := unb.index(r,"$OFFER$");
                    unb.replace_slice(r,index,index - 1 +
                                      unb.length(unb.to_unbounded_string("$OFFER$")),
                                      trim(unb.to_string(hr),ada.strings.both) &
                                      "<form method=post action=""/accept"">" & ascii.lf &
                                      "<input type=""hidden"" name=""offer"" value=""" &
                                      trim(natural'image(offer.offer.offer_id),ada.strings.both) &
                                      """>" & ascii.lf &
                                      "<input type=""hidden"" name=""offerer"" value=""" &
                                      trim(unb.to_string(offer.offer.offerer_id),ada.strings.both) &
                                      """>" & ascii.lf &
                                      "<input type=""hidden"" name=""session"" value=""" &
                                      trim(integer'image(auth.get_id(s)),ada.strings.both) &
                                      """>" & ascii.lf & unb.to_string(color_left) &
                                      "Oferta: " &
                                      trim(unb.to_string(offer.offer.offerer_id),ada.strings.both) &
                                      ", " & utl.image(offer.tmstamp) & 
                                      ", " & trim(natural'image(offer.offer.stock_price),ada.strings.both) &
                                      " (precio/unidad), " &
                                      trim(natural'image(offer.offer.quantity),ada.strings.both) &
                                      ", " & unb.to_string(t) & ascii.lf & unb.to_string(color_right) &
                                      "<input type=""submit"" value=""Aceptar""><br>" & ascii.lf &
                                      "<ul>$ACCEPT$</ul>" & ascii.lf &
                                      "</form>" & ascii.lf &
                                      "$OFFER$");
                else    -- outdated offer, we cant accept this!
                    -- generate html
                    index := unb.index(r,"$OFFER$");
                    unb.replace_slice(r,index,index - 1 +
                                      unb.length(unb.to_unbounded_string("$OFFER$")),
                                      trim(unb.to_string(hr),ada.strings.both) &
                                      unb.to_string(color_left) &
                                      "Oferta: " &
                                      trim(unb.to_string(offer.offer.offerer_id),ada.strings.both) &
                                      ", " & utl.image(offer.tmstamp) & 
                                      ", " & trim(natural'image(offer.offer.stock_price),ada.strings.both) &
                                      " (precio/unidad), " &
                                      trim(natural'image(offer.offer.quantity),ada.strings.both) &
                                      ", " & unb.to_string(t) & ascii.lf & unb.to_string(color_right) &
                                      "<ul>$ACCEPT$</ul>" & ascii.lf &
                                      "$OFFER$");
                end if;

                -- print accepts
                if not al.is_empty(offer.accepts) then
                    aref := al.last(offer.accepts);
                end if;
                
                while aref /= al.null_reference loop
                    accep := al.get_element(offer.accepts,aref);
                    -- check if we've done this accept, and put in red if so
                    if accep.acceptor_id = data.login then
                        color_left := unb.to_unbounded_string("<font color=""red"">");
                        color_right := unb.to_unbounded_string("</font>");
                    else
                        color_left := unb.to_unbounded_string("");
                        color_right := unb.to_unbounded_string("");
                    end if;
                    -- generate html
                    index := unb.index(r,"$ACCEPT$");
                    unb.replace_slice(r,index,index - 1 +
                                      unb.length(unb.to_unbounded_string("$ACCEPT$")),
                                      "<li>" & unb.to_string(color_left) & "Aceptada por: " &
                                      trim(unb.to_string(accep.acceptor_id),ada.strings.both) &
                                      ", " & utl.image(accep.tmstamp) & unb.to_string(color_right) &
                                      "</li>" & ascii.lf & "$ACCEPT$");
                    al.rewind(offer.accepts,aref);
                end loop;

                index := unb.index(r,"$ACCEPT$");
                if offer.commited_to /= unb.null_unbounded_string then
                    -- substitute trailing $ACCEPT$ by a commit
                    unb.replace_slice(r,index,index - 1 +
                                      unb.length(unb.to_unbounded_string("$ACCEPT$")),
                                      "<li><b>Comprometida para: " &
                                      trim(unb.to_string(offer.commited_to),ada.strings.both) &
                                      ", " & utl.image(offer.commited_at) & "</b></li>" & ascii.lf);
                else
                    -- delete last
                    r := unb.delete(r,index,index - 1 +
                                    unb.length(unb.to_unbounded_string("$ACCEPT$")));
                end if;

                -- go next!
                ol.rewind(offers,oref);
            end loop;

            -- delete last
            index := unb.index(r,"$OFFER$");
            r := unb.delete(r,index,index - 1 +
                            unb.length(unb.to_unbounded_string("$OFFER$")));

            http_common.header_line(str,http_common.CONTENTLEN,
                                    unb.to_unbounded_string(integer'image(
                                    unb.length(r))));
            unb.append(str,ascii.cr & ascii.lf);
            unb.append(str,r);

            return unb.to_string(str);
        end if;

    exception
        when e: auth.badsession =>
            return build_response(ERROR);
        when e: http_common.unknown_var =>
            return build_response(ERROR);
        when e: others =>
            utl.error("http: unexpected error: " & ada.exceptions.exception_message(e));
            return build_response(ERROR);
    end;

    procedure bind_multicast(addr: udp.end_point_type) is
    begin
        multicast_main_addr := addr;
        udp.bind(addr,hdl.main_handler'access);
    end;

end broker_control;
