-- Broker message processing implementation
-- File name: broker_messages.adb
-- Author: Jaime Prez Crespo <japecre@gsyc.escet.urjc.es>
-- Last modification: June 16, 2005
--
-- Copyright (c) 2005, Jaime Prez Crespo
-- All rights reserved.
-- 
-- Redistribution and use in source and binary forms, with or without modification,
-- are permitted provided that the following conditions are met:
-- - Redistributions of source code must retain the above copyright notice, this list
--   of conditions and the following disclaimer.
-- - Redistributions in binary form must reproduce the above copyright notice, this list
--   of conditions and the following disclaimer in the documentation and/or other materials
--   provided with the distribution.
-- - Neither the name of the Rey Juan Carlos University nor the names of its contributors
--   may be used to endorse or promote products derived from this software without specific
--   prior written permission.
-- 
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
-- OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
-- AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
-- CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
-- DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-- DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
-- IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-- OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

with ada.strings.unbounded;
with lower_layer_udp;
with utils;

package body broker_messages is

    package unb renames ada.strings.unbounded;
    package udp renames lower_layer_udp;
    package utl renames utils;
    use al; -- accept lists
    use ol; -- offer lists

    -- message processing
    procedure process(a_message: in message_offer) is
        o: offer_queue;
        ho: history_queue(t => MOFFER);
        b: broker_queue;
        r: bl.reference;
    begin
        if current_entity /= GONG then
            o.offer := a_message;
            o.tmstamp := ada.calendar.clock;

            if current_entity = OBSERVER then
                ho.offer := a_message;
                ho.tmstamp := o.tmstamp;
                hl.insert_tail(history,ho);

                -- new broker?
                begin
                    r := search(brokers,a_message.offerer_id);
                    b := bl.get_element(brokers,r);
                exception
                    when ebadbroker =>
                        -- this is a new broker, so register it
                        b.id := a_message.offerer_id;
                        bl.insert_tail(brokers,b);
                        r := bl.last(brokers);
                        utl.debug("process: new broker " & unb.to_string(b.id) &
                                  " inserted" & ascii.lf);
                end;

                -- check if user has enough room to perform the operation, and
                -- if so, accept it
                if (o.offer.kind = SELL and then o.offer.quantity <= b.vstock) or
                   (o.offer.kind = BUY and then 
                    o.offer.quantity * o.offer.stock_price <= b.vmoney) then
                    -- reflect changes on broker data
                    if o.offer.kind = BUY then
                        b.vmoney := b.vmoney - o.offer.quantity * o.offer.stock_price;
                        -- you can't sell stocks you don't have yet!
                        --b.vstock := b.vstock + o.offer.quantity; 
                    else
                        -- sell the rest of stock, but can't use money you're not
                        -- sure to have!
                        --b.vmoney := b.vmoney + o.offer.quantity * o.offer.stock_price;
                        b.vstock := b.vstock - o.offer.quantity;
                    end if;

                    -- update broker accounting and accept offer
                    bl.replace(brokers,b,r);
                    ol.insert_tail(offers,o);
                end if;
            else    -- broker
                ol.insert_tail(offers,o);
            end if;
        end if;
    exception
        when ol.storage_exhausted =>
            utl.error("process: storage exhausted, cannot record offer");
    end;

    procedure process(a_message: in message_accept) is
        r: ol.reference;
        s: al.reference;
        t: bl.reference;
        o: offer_queue;
        a: accept_queue;
        bo: broker_queue;
        ba: broker_queue;
        ha: history_queue(t => MACCEPT);
        hc: history_queue(t => MCOMMIT);
        c: message_commit;
        buf: aliased udp.buffer_type(1024);
    begin
        if current_entity /= GONG then
            -- find offer
            begin
                r := search(a_message.offer_id,a_message.offerer_id);
                o := ol.get_element(offers,r);
                -- build accept
                a.tmstamp := ada.calendar.clock;
                a.acceptor_id := a_message.acceptor_id;
            exception
                when e: ebadoffer =>
                    utl.error("process: accept received for inexistent or old offer");
            end;

            if current_entity = OBSERVER then
                ha.tmstamp := ada.calendar.clock;
                ha.accep := a_message;
                hl.insert_tail(history,ha);
                if r /= ol.null_reference then
                    begin
                        -- find broker who accepted the offer (or register it)
                        t := search(brokers,a_message.acceptor_id);
                        ba := bl.get_element(brokers,t);
                    exception
                        when ebadbroker =>
                            -- this is a new broker, so register it
                            ba.id := a_message.acceptor_id;
                            bl.insert_tail(brokers,ba);
                            t := bl.last(brokers);
                            utl.debug("process: new broker " & unb.to_string(ba.id) &
                                      " inserted" & ascii.lf);
                    end;

                    -- check if we must do a commit
                    if al.is_empty(o.accepts) and then
                       ((o.offer.kind = BUY and then o.offer.quantity <= ba.vstock) or
                       (o.offer.kind = SELL and then
                        (o.offer.stock_price * o.offer.quantity <= ba.vmoney))) then
                        -- the last was also the first and unique accept, the
                        -- one we've inserted here, so deserves a commit
                        utl.debug("process: commiting offer" & ascii.lf);

                        -- update acceptor accounting
                        if o.offer.kind = BUY then
                            ba.money := ba.money + o.offer.stock_price * o.offer.quantity;
                            ba.vmoney := ba.vmoney + o.offer.stock_price * o.offer.quantity;
                            ba.stock := ba.stock - o.offer.quantity;
                            ba.vstock := ba.vstock - o.offer.quantity;
                        else
                            ba.money := ba.money - o.offer.stock_price * o.offer.quantity;
                            ba.vmoney := ba.vmoney - o.offer.stock_price * o.offer.quantity;
                            ba.stock := ba.stock + o.offer.quantity;
                            ba.vstock := ba.vstock + o.offer.quantity;
                        end if;
                        bl.replace(brokers,ba,t);

                        -- find broker who did the offer. This won't fail, the
                        -- offer exists, so the offerer is registered!
                        t := search(brokers,a_message.offerer_id);
                        ba := bl.get_element(brokers,t);

                        -- update his accounting
                        if o.offer.kind = BUY then
                            ba.money := ba.money - o.offer.stock_price * o.offer.quantity;
                            ba.vmoney := ba.vmoney - o.offer.stock_price * o.offer.quantity;
                            ba.stock := ba.stock + o.offer.quantity;
                            ba.vstock := ba.vstock + o.offer.quantity;
                        else
                            ba.money := ba.money + o.offer.stock_price * o.offer.quantity;
                            ba.vmoney := ba.vmoney + o.offer.stock_price * o.offer.quantity;
                            ba.stock := ba.stock - o.offer.quantity;
                            ba.vstock := ba.vstock - o.offer.quantity;
                        end if;
                        bl.replace(brokers,ba,t);

                        c.offer_id := o.offer.offer_id;
                        c.offerer_id := o.offer.offerer_id;
                        c.acceptor_id := a_message.acceptor_id;

                        -- pack and send
                        udp.reset(buf);
                        message'class'output(buf'access,message'class(c));
                        udp.send(observer_data.multicast_main_addr,buf'access);

                        -- now, store the commit
                        hc.tmstamp := ada.calendar.clock;
                        hc.offer_id := c.offer_id;
                        hc.offerer_id := c.offerer_id;
                        hc.acceptor_id := c.acceptor_id;
                        hl.insert_tail(history,hc);

                        -- insert accept on offer accept queue
                        al.insert_tail(o.accepts,a);
                        -- update offer
                        ol.replace(offers,o,r);
                    end if;
                end if;
            elsif r /= ol.null_reference then
                -- insert accept on offer accept queue
                al.insert_tail(o.accepts,a);
                -- update offer
                ol.replace(offers,o,r);
            end if;
        end if;
    end;

    procedure process(a_message: in message_commit) is
        use type unb.unbounded_string;
        r: ol.reference;
        s: al.reference;
        o: offer_queue;
        a: accept_queue;
        session: auth.session;
        d: auth.auth_info_addr;
    begin
        if current_entity = BROKER then
            -- verify commit data and update structures
            r := search(a_message.offer_id,a_message.offerer_id);
            o := ol.get_element(offers,r);
            s := search(o.accepts,a_message.acceptor_id);
            a := al.get_element(o.accepts,s);

            o.commited_to := a_message.acceptor_id;
            o.commited_at := ada.calendar.clock;
            ol.replace(offers,o,r);

            -- does this offer belong to a local user?
            begin
                session := auth.search(o.offer.offerer_id);
                d := auth.get_info(session);
                if o.offer.kind = BUY then
                    d.stock := d.stock + o.offer.quantity;
                else
                    d.stock := d.stock - o.offer.quantity;
                end if;
            exception
                when e: auth.badsession =>
                    -- commit granted to foreign user, silently ignore
                    null;
            end;

            -- search for local users who must be reimbursed
            s := al.first(o.accepts);
            while s /= al.null_reference loop
                a := al.get_element(o.accepts,s);
                begin
                    session := auth.search(a.acceptor_id);
                    d := auth.get_info(session);
                    -- a local user accepted this offer, is the commit for him?
                    if d.login /= a_message.acceptor_id then
                        if o.offer.kind = BUY then
                            d.vstock := d.vstock - o.offer.quantity;
                            d.money := d.money + o.offer.stock_price;
                        else
                            d.vstock := d.vstock + o.offer.quantity;
                            d.money := d.money - o.offer.stock_price;
                        end if;
                    else   
                        -- a local user who accepted this offer got the commit,
                        -- update his real stock
                        if o.offer.kind = BUY then
                            d.stock := d.stock - o.offer.quantity;
                        else
                            d.stock := d.stock + o.offer.quantity;
                        end if;
                    end if;
                exception
                    when e: auth.badsession =>
                        -- accept for non local user, silently ignore and
                        -- continue looking for more accepts
                        null;
                end;
                al.forward(o.accepts,s);
            end loop;
        end if;

    exception
        when e: ebadoffer =>
            -- outdated offer, silently ignore
            null;
        when e: ebadaccep =>
            -- unknown accept, silently ignore
            null;
    end;

    procedure process(a_message: in message_gong) is
        use type bl.reference;
        hg: history_queue(t => MGONG);
        r: bl.reference := bl.null_reference;
        b: broker_queue;
    begin
        if current_entity /= GONG then
            broker_data.round := a_message.round_id;
            utl.debug("gong received, new round started" & ascii.lf);
            -- IMPORTANT: brokers will obtain their per round benefits
            -- when they DO SOME OPERATION. Please note this outdates
            -- information on brokers ONLY when it is NOT NEEDED. It
            -- must be done like this due to the multiple users a broker
            -- can handle and the generic authentication facilities
            sync;     -- synchronize broker accounts
            purge;    -- purge the list of outdated offers
            if current_entity = OBSERVER then
                hg.tmstamp := ada.calendar.clock;
                hg.round := a_message.round_id;
                hl.insert_tail(history,hg);

                -- update broker accounts
                if not bl.is_empty(brokers) then
                    r := bl.first(brokers);
                end if;

                while r /= bl.null_reference loop
                    b := bl.get_element(brokers,r);
                    -- we can do this because accounts are synced
                    b.money := b.money + b.stock;
                    -- synchronize accounting (undo offers and accepts)
                    b.vmoney := b.money;
                    b.vstock := b.stock;
                    bl.replace(brokers,b,r);
                    bl.forward(brokers,r);
                end loop;
            end if;
        end if;
    end;

    -- lists processing

    -- debugging related functions ---------------------------------------------
    -- NOT IMPLEMENTED!
    function accept_image(q: accept_queue) return string is
    begin
        return unb.to_string(q.acceptor_id);
    end;

    function offer_image(q: offer_queue) return string is
        str: unb.unbounded_string;
    begin
        return unb.to_string(str);
    end;
    
    function history_image(q: history_queue) return string is
        str: unb.unbounded_string;
    begin
        return unb.to_string(str);
    end;
    ----------------------------------------------------------------------------

    function search(offer: natural;
                   offerer: asu.unbounded_string) return ol.reference is
        use type unb.unbounded_string;
        r: ol.reference;
        e: offer_queue;
    begin
        if not ol.is_empty(offers) then
            r := ol.first(offers);
        end if;

        while r /= ol.null_reference loop
            e := ol.get_element(offers,r);
            if e.offer.offer_id = offer and e.offer.offerer_id = offerer then
                return r;
            end if;
            ol.forward(offers,r);
        end loop;

        raise ebadoffer;
    end;

    function search(l: al.list; id: asu.unbounded_string) return al.reference is
        use type unb.unbounded_string;
        r: al.reference;
        e: accept_queue;
    begin
        if not al.is_empty(l) then
            r := al.first(l);
        end if;

        while r /= al.null_reference loop
            e := al.get_element(l,r);
            if e.acceptor_id = id then
                return r;
            end if;
            al.forward(l,r);
        end loop;

        raise ebadaccep;
    end;

    procedure purge is
        r,s: ol.reference;
        e: offer_queue;
        offset: natural := 0;
    begin
        if current_entity = BROKER then
            -- this offset determines the number of rounds beyond the current
            -- a offer is kept
            offset := 1;
        end if;

        if not ol.is_empty(offers) then
            r := ol.first(offers);
        else
            return;
        end if;

        while not ol.is_last(offers,r) loop
            e := ol.get_element(offers,r);
            s := r;
            ol.forward(offers,r);
            if e.offer.round_id < broker_data.round - offset then
                al.finalize(e.accepts);
                ol.remove(offers,s);
            end if;
        end loop;

        -- last one
        e := ol.get_element(offers,r);
        if e.offer.round_id < broker_data.round - 1 then
            ol.remove(offers,r);
        end if;
    end;

    procedure sync is
        r: ol.reference;
        e: offer_queue;
        s: auth.session;
        d: auth.auth_info_addr;
    begin
        if not ol.is_empty(offers) then
            r := ol.first(offers);
        else
            return;
        end if;

        -- update accounting for outdated offers
        while r /= ol.null_reference loop
            e := ol.get_element(offers,r);
            if e.offer.round_id < (broker_data.round - 1) and al.is_empty(e.accepts) then
                -- our offer was not accepted, restore money/stock
                -- as only virtual stock is modified, we need only to update this
                begin
                    s := auth.search(e.offer.offerer_id);
                    d := auth.get_info(s);
                    if e.offer.kind = BUY then
                        d.vstock := d.vstock - e.offer.quantity;
                        d.money := d.money + e.offer.stock_price;
                    else
                        d.vstock := d.vstock + e.offer.quantity;
                        d.money := d.money - e.offer.stock_price;
                    end if;
                exception
                    when e: auth.badsession =>
                        -- remote user, this is expected, just ignore it
                        null;
                end;
            end if;
            ol.forward(offers,r);
        end loop;

    end;

end broker_messages;
