-- Broker message processing specification
-- File name: broker_messages.ads
-- Author: Jaime Prez Crespo <japecre@gsyc.escet.urjc.es>
-- Last modification: June 15, 2005
--
-- This package uses code borrowed from
-- http://gsyc.escet.urjc.es/docencia/asignaturas/redes-II/practicas-2003-04/broker_messages.ads
-- Code related to HTTP:
-- Copyright (c) 2005, Jaime Prez Crespo
-- Everything else is property of Luis Lpez Fernndez <llopez@gsyc.escet.urjc.es>.
-- All rights reserved.
--
-- Redistribution and use in source and binary forms, with or without modification,
-- are permitted provided that the following conditions are met:
-- - Redistributions of source code must retain the above copyright notice, this list
--   of conditions and the following disclaimer.
-- - Redistributions in binary form must reproduce the above copyright notice, this list
--   of conditions and the following disclaimer in the documentation and/or other materials
--   provided with the distribution.
-- - Neither the name of the Rey Juan Carlos University nor the names of its contributors
--   may be used to endorse or promote products derived from this software without specific
--   prior written permission.
-- 
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
-- OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
-- AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
-- CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
-- DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-- DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
-- IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-- OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

with ada.strings.unbounded;
with ada.calendar;
with lists_generic;
with broker_data;
with observer_data;
with mod_auth;

package broker_messages is

    package asu renames ada.strings.unbounded;
    use broker_data;
    use observer_data;

    -------------------------------------------------------------------
    -- root of message hierarchy
    -------------------------------------------------------------------

    type message is abstract tagged null record;

    procedure process(a_message: in message) is abstract;

    -------------------------------------------------------------------
    -- offer
    -------------------------------------------------------------------

    type offer_kind is (BUY, SELL);

    type message_offer is new message with record
        kind        : offer_kind;
        quantity    : natural;
        stock_price : natural;
        offer_id    : natural;
        offerer_id  : asu.unbounded_string;
        round_id    : natural;
    end record;

    procedure process(a_message: in message_offer);

    -------------------------------------------------------------------
    -- accept
    -------------------------------------------------------------------

    type message_accept is new message with record
        offer_id    : natural;
        offerer_id  : asu.unbounded_string;
        acceptor_id : asu.unbounded_string;
    end record;

    procedure process(a_message: in message_accept);

    -------------------------------------------------------------------
    -- commit
    -------------------------------------------------------------------

    type message_commit is new message with record
        offer_id    : natural;
        offerer_id  : asu.unbounded_string;
        acceptor_id : asu.unbounded_string;
    end record;

    procedure process(a_message: in message_commit);

    -------------------------------------------------------------------
    -- gong
    -------------------------------------------------------------------

    type message_gong is new message with record
        round_id    : natural;
    end record;
    
    procedure process(a_message: in message_gong);

    -------------------------------------------------------------------
    -- exceptions
    -------------------------------------------------------------------

    ebadoffer: exception;
    ebadaccep: exception;

    -------------------------------------------------------------------
    -- http messages
    -------------------------------------------------------------------

    type response is (ROOT, REDIRECT, OPTIONS, ERROR);
    type entity is (BROKER, GONG, OBSERVER);
    current_entity: entity;

    -------------------------------------------------------------------
    -- message queues
    -------------------------------------------------------------------

    -- accept lists
    type accept_queue is record
        acceptor_id: asu.unbounded_string;
        tmstamp: ada.calendar.time;
    end record;
    function accept_image(q: accept_queue) return string;
    package al is new lists_generic(accept_queue,accept_image);

    null_commit: constant message_commit := (0,asu.to_unbounded_string(""),
                                             asu.to_unbounded_string(""));
    -- offer list
    type offer_queue is record
        tmstamp: ada.calendar.time;
        offer: message_offer;
        accepts: al.list := al.null_list;
        commited_to: asu.unbounded_string := asu.null_unbounded_string;
        commited_at: ada.calendar.time;
    end record;
    function offer_image(q: offer_queue) return string;
    package ol is new lists_generic(offer_queue,offer_image);

    -- history list
    MOFFER:  constant := 1;
    MACCEPT: constant := 2;
    MGONG:   constant := 3;
    MCOMMIT: constant := 4;
    subtype message_type is integer range MOFFER..MCOMMIT;
    type history_queue(t: message_type := MOFFER) is record
        tmstamp: ada.calendar.time;
        case t is
        when MOFFER =>
            offer: message_offer;
        when MACCEPT =>
            accep: message_accept;
        when MGONG =>
            round: integer;
        when MCOMMIT =>
            offer_id: natural;
            offerer_id: asu.unbounded_string;
            acceptor_id: asu.unbounded_string;
        when others =>
            null;
        end case;
    end record;
    function history_image(q: history_queue) return string;
    package hl is new lists_generic(history_queue,history_image);

    offers: ol.list := ol.null_list;
    history: hl.list := hl.null_list;

    function search(offer: natural;
                    offerer: asu.unbounded_string) return ol.reference;
    -- search for an offer and returns a pointer to it. May raise ebadoffer if
    -- no offer found

    function search(l: al.list; id: asu.unbounded_string) return al.reference;
    -- search for an accept and returns a pointer to it. May raise ebadaccep if
    -- no accep found

private

    procedure purge;
    -- search the offers list for to rounds old offers, and remove them
   
    procedure sync;
    -- synchonize broker accounting 

end broker_messages;
