-- HTTP_Server package implementation
-- File name: http_request.adb
-- Author: Jaime Prez Crespo <japecre@gsyc.escet.urjc.es>
-- Last modification: May 16, 2005
--
-- Copyright (c) 2005, Jaime Prez Crespo
-- All rights reserved.
-- 
-- Redistribution and use in source and binary forms, with or without modification,
-- are permitted provided that the following conditions are met:
-- - Redistributions of source code must retain the above copyright notice, this list
--   of conditions and the following disclaimer.
-- - Redistributions in binary form must reproduce the above copyright notice, this list
--   of conditions and the following disclaimer in the documentation and/or other materials
--   provided with the distribution.
-- - Neither the name of the Rey Juan Carlos University nor the names of its contributors
--   may be used to endorse or promote products derived from this software without specific
--   prior written permission.
-- 
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
-- OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
-- AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
-- CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
-- DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-- DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
-- IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-- OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
--
-- This package uses data structures and parsing code borrowed from
-- http://gsyc.escet.urjc.es/docencia/asignaturas/ral-04-05/practicas/petition_analysis.ads and
-- http://gsyc.escet.urjc.es/docencia/asignaturas/ral-04-05/practicas/petition_analysis.adb, so
-- the vast majority of code is property of its authors:
-- Antonio Fernndez Anta <anto@gsyc.escet.urjc.es> and
-- Luis Rodero Merino <lrodero@gsyc.escet.urjc.es>

with ada.strings.maps;
with ada.exceptions;
with ada.characters.handling;

package body http_request is

    package Maps renames Ada.Strings.Maps;
    package Exceptions renames Ada.Exceptions;
    use ada.strings.unbounded;
    use ada.characters.handling;

    procedure findstr(str: in unbounded.unbounded_string;
                      token: in character;
                      lower,upper: in integer;
                      exists: out boolean;
                      pos: out integer) is
        char: character;
    begin
        exists := false;
        pos := lower;
        for i in lower..upper loop
            char := unbounded.element(str,i);
            if char = token then
                exists := true;
                pos := i;
                exit;
            end if;
        end loop;
    end findstr;

    function break_uri(url: in unbounded.unbounded_string) return turi is
        protocol, host, port, path: unbounded.unbounded_string;
        position,pos_bot,pos_top: integer;
        tmp,head: unbounded.unbounded_string;
        hasprotocol: boolean := false;
        hasport: boolean := false;
        haspath:boolean := false;
        final: integer;
        crlf: string(1..2);
        uri: turi;
    begin

        port := unbounded.to_unbounded_string("80");
        path := unbounded.to_unbounded_string("/");
        host := unbounded.to_unbounded_string("");
        protocol := unbounded.to_unbounded_string("http");

        crlf(1) := ascii.cr;
        crlf(2) := ascii.lf;
        tmp := url;
        final := unbounded.length(tmp);

        -- protocol defined?
        findstr(tmp,':',1,final,hasprotocol,position);

        if (hasprotocol) then
            -- is there "://"?
            if (unbounded.to_unbounded_string(unbounded.slice(tmp,5,7)) /= unbounded.to_unbounded_string("://")) then
                raise bad_syntax;
            end if;

            -- save protocol
            protocol := unbounded.to_unbounded_string(unbounded.slice (tmp,1,position-1));

            -- look for a port
            findstr(tmp,':',position+1,final,hasport,pos_bot);

            if (hasport) then
                findstr(tmp,'/',position+3,final,haspath,pos_top);

                if (pos_bot = pos_top - 1) then
                    raise bad_syntax;
                end if;

                if (haspath) then
                    port := unbounded.to_unbounded_string(unbounded.slice(tmp,pos_bot +1,pos_top -1));
                    host := unbounded.to_unbounded_string(unbounded.slice(tmp,position +3,pos_bot-1));
                    path := unbounded.to_unbounded_string(unbounded.slice(tmp,pos_top,final));
                else
                    host := unbounded.to_unbounded_string(unbounded.slice(tmp,position +3,final));
                end if;
            else
                findstr(tmp,'/',position +3,final,haspath,pos_top);
                if (haspath) then
                    path := unbounded.to_unbounded_string(unbounded.slice(tmp,pos_top,final));
                    host := unbounded.to_unbounded_string(unbounded.slice(tmp,position +3,pos_top -1));
                else
                    host := unbounded.to_unbounded_string(unbounded.slice(tmp,position +3,final));
                end if;
            end if;
        else
            if (unbounded.element(tmp,1) = '/') then
                path := unbounded.to_unbounded_string(unbounded.slice(tmp,1,final));
            else
                raise bad_syntax;
            end if;
        end if;

        -- fill in structure
        uri.protocol := protocol;
        uri.host := host;
        uri.port := port;
        uri.path := path;

        return uri;
    end break_uri;

    -- analyze header fields
    procedure analyze_field(field_line: in out unbounded.unbounded_string;
                            field: in out theader) is
        colon_index: natural;
    begin                                
        colon_index := unbounded.index(field_line, ":");
        if colon_index <= 1 then
            exceptions.raise_exception(bad_syntax'identity, "field without colon");
        end if;
        field.name := unbounded.trim(unbounded.to_unbounded_string(to_lower(
                       unbounded.slice(field_line,1,colon_index-1))), ada.strings.both);
        field.value := unbounded.trim(unbounded.to_unbounded_string(
                        unbounded.slice(field_line,colon_index+1,unbounded.length(field_line))),
                                        ada.strings.both);
    end analyze_field;

    -- analyze first line of a request
    procedure analyze_first_line(first_line: in out unbounded.unbounded_string;
                                 petition: in out request) is
        blank_index: natural;
        method_string: unbounded.unbounded_string;
    begin
        -- getting method
        blank_index := unbounded.index(first_line, maps.to_set(' '));
        if blank_index < 3 then
            exceptions.raise_exception(bad_syntax'identity, "no command in first line");
        end if;
        method_string := unbounded.to_unbounded_string(to_lower(unbounded.slice(first_line, 1, blank_index - 1)));
        if method_string = "options" then
            petition.method := OPTIONS;
        elsif method_string = "get" then
            petition.method := GET;
        elsif  method_string = "head" then
            petition.method := HEAD;
        elsif  method_string = "post" then
            petition.method := POST;
        elsif  method_string = "put" then
            petition.method := PUT;
        elsif  method_string = "delete" then
            petition.method := DELETE;
        elsif  method_string = "trace" then
            petition.method := TRACE;
        elsif  method_string = "connect" then
            petition.method := CONNECT;
        else    
            exceptions.raise_exception(unknown_method'identity, "unknown method " &
                                       unbounded.to_string(method_string));
        end if;   
        -- removing method
        unbounded.delete(first_line,1,blank_index);
        
        -- getting uri
        blank_index := unbounded.index(first_line, maps.to_set(' '));
        if blank_index < 1 then
            exceptions.raise_exception(bad_syntax'identity, "no uri defined");
        end if;
        petition.uri := break_uri(unbounded.to_unbounded_string(unbounded.slice(first_line, 1, blank_index - 1)));
        -- removing uri and lowering case
        unbounded.delete(first_line,1,blank_index);
        first_line := unbounded.to_unbounded_string(to_lower(unbounded.to_string(first_line)));

        -- getting version
        if first_line = "http/1.0" then
            petition.proto := HTTP10;
        elsif first_line = "http/1.1" then
            petition.proto := HTTP11;
            petition.close := false;
         else
            exceptions.raise_exception(unknown_proto'identity, "unknown protocol " &
                                       unbounded.to_string(first_line));
            --exceptions.raise_exception(bad_syntax'identity, "unknown version " & unbounded.to_string(first_line));
        end if;
    exception
        when e: bad_syntax =>
            exceptions.raise_exception(bad_syntax'identity, "uri not well formed: " & exceptions.exception_message(e));
        when e:  others =>
            exceptions.raise_exception(bad_syntax'identity, "unexpected error, message is: " & exceptions.exception_message(e));
    end analyze_first_line;

    -- analyze a complete request
    procedure analyze_request(buf: in out unbounded.unbounded_string;
                              req: in out request) is
        first_line: unbounded.unbounded_string;
        field_line: unbounded.unbounded_string;
        number_of_fields: natural;
        fields_array_access: theaders;
        cr_lf_index: natural;
    begin
        -- analyzing first line
        cr_lf_index := unbounded.index(buf, ascii.cr & ascii.lf & "");
        
        if cr_lf_index = 0 then
            exceptions.raise_exception(bad_syntax'identity, "no cr or lf char found");
        elsif cr_lf_index = 1 then    
            exceptions.raise_exception(bad_syntax'identity, "first line empty");
        end if;
        
        first_line := unbounded.to_unbounded_string(unbounded.slice(buf, 1, cr_lf_index-1));
        
        analyze_first_line(first_line, req);

        -- removing first line
        unbounded.delete(buf, 1, cr_lf_index + 1);

        -- analyzing fields
        number_of_fields := unbounded.count(buf,ascii.cr & ascii.lf & "") - 1;
        req.nheaders := number_of_fields;
        fields_array_access := new theaders_record(number_of_fields);
        req.headers := fields_array_access;
        for loop_index in 1..number_of_fields loop
            cr_lf_index := unbounded.index(buf, ascii.cr & ascii.lf & "");
            field_line := unbounded.to_unbounded_string(unbounded.slice(buf, 1,cr_lf_index-1));
            analyze_field(field_line, fields_array_access.headers(loop_index));
            -- check if connection closing required
            if unbounded.to_string(fields_array_access.headers(loop_index).name) = "connection" and then
                to_lower(unbounded.to_string(fields_array_access.headers(loop_index).value)) = "keep-alive" then
                req.close := false;
            end if;
            -- check if body
            if unbounded.to_string(fields_array_access.headers(loop_index).name) = "content-length" then
                req.datalen := integer'value(unbounded.to_string(fields_array_access.headers(loop_index).value));
            end if;
            -- removing field 
            unbounded.delete(buf, 1, cr_lf_index + 1);
        end loop;
    end;

end http_request;

