-- Generic authentication package specification
-- Follow these steps to use it:
-- - Create a new session
-- - Associate your custom data to it (including username and password)
-- - Authenticate it
-- Associations are locally stored. Use the appropiate functions to retrieve
-- a session from a session id.
-- File name: mod_auth.ads
-- Author: Jaime Prez Crespo <japecre@gsyc.escet.urjc.es>
-- Last modification: Jun 15, 2005
--
-- Copyright (c) 2005, Jaime Prez Crespo
-- All rights reserved.
-- 
-- Redistribution and use in source and binary forms, with or without modification,
-- are permitted provided that the following conditions are met:
-- - Redistributions of source code must retain the above copyright notice, this list
--   of conditions and the following disclaimer.
-- - Redistributions in binary form must reproduce the above copyright notice, this list
--   of conditions and the following disclaimer in the documentation and/or other materials
--   provided with the distribution.
-- - Neither the name of the Rey Juan Carlos University nor the names of its contributors
--   may be used to endorse or promote products derived from this software without specific
--   prior written permission.
-- 
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
-- OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
-- AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
-- CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
-- DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-- DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
-- IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-- OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

with ada.strings.unbounded;
with unchecked_deallocation;
with mod_auth_basic;

generic
    type auth_info is private;
    -- type to store user data associated with authentication

    with function get_auth_info(info: auth_info) return mod_auth_basic.auth_basic_info;
    -- this package provides generic functionality to store user-defined data
    -- associated with sessions. It also provides basic authentication schemes,
    -- so it needs some data to be well known. This function should implement the
    -- way to retrieve this data (an username and a password) from the custom
    -- structure auth_info.

package mod_auth is

    enotmem: exception; -- not enough memory to perform requested action
    badsession: exception;  -- invalid session specified

    type session is private;
    type auth_info_addr is access auth_info;

    function new_session return session;
    -- creates a new authentication session with its unique identifier
    -- may raise enotmem if not enough memory to allocate a new session

    procedure associate(s: in out session; data: auth_info_addr);
    -- create an association for an existing session

    function authenticate(s: session) return boolean;
    -- authenticates user basic info for the current session

    function validate(id: integer) return session;
    -- search and retrieve a session by its id
    -- may raise badsession if the session id is unknown

    function search(usr: ada.strings.unbounded.unbounded_string) return session;
    -- search and retrieve a session by its username
    -- may raise badsession if the session id is unknown

    function get_id(s: session) return integer;
    -- retrieves session unique identifier

    function get_info(s: session) return auth_info_addr;
    -- retrieves session custom private information

    procedure finnish(s: session);
    -- ends the current authentication session
    -- may raise badsession if the session is invalid

private

    -- types and variables for handling sessions
    type session is record
        id:   integer;
        priv: auth_info_addr;
    end record;

    MAXSESSIONS: constant := 100;
    sessions: array(1..MAXSESSIONS) of session;
    nsessions: integer range 0..MAXSESSIONS := 0;

    -- types and variables for handling user authentication
    MAXUSERS: constant := 100;
    clasic_auth: array(1..MAXUSERS) of mod_auth_basic.auth_basic_info;

end mod_auth;
