-- Observer control implementation
-- File name: observer_control.ads
-- Author: Jaime Prez Crespo <japecre@gsyc.escet.urjc.es>
-- Last modification: June 15, 2005
--
-- Copyright (c) 2005, Jaime Prez Crespo
-- All rights reserved.
-- 
-- Redistribution and use in source and binary forms, with or without modification,
-- are permitted provided that the following conditions are met:
-- - Redistributions of source code must retain the above copyright notice, this list
--   of conditions and the following disclaimer.
-- - Redistributions in binary form must reproduce the above copyright notice, this list
--   of conditions and the following disclaimer in the documentation and/or other materials
--   provided with the distribution.
-- - Neither the name of the Rey Juan Carlos University nor the names of its contributors
--   may be used to endorse or promote products derived from this software without specific
--   prior written permission.
-- 
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
-- OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
-- AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
-- CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
-- DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-- DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
-- IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-- OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

with ada.text_io;
with ada.numerics.discrete_random;
with ada.strings.unbounded;
with ada.exceptions;
with ada.characters.latin_1;
with ada.strings.fixed;
with ada.calendar;
with broker_messages;
with observer_data;
with handlers;
with http_common;
with utils;

package body observer_control is

    package udp renames lower_layer_udp;
    package unb renames ada.strings.unbounded;
    package utl renames utils;
    package hdl renames handlers;
    use ada.strings.fixed;
    use broker_messages;
    use observer_data;

    function build_response(r: response) return string is
        str: unb.unbounded_string;
    begin
        case r is
        when ROOT =>
            str := http_common.status_line(http_common.OK,http_common.HTTP11);
            http_common.header_line(str,http_common.SERVER,observer_identification);
        when ERROR =>
            str := http_common.status_line(http_common.OK,http_common.HTTP11);
            http_common.header_line(str,http_common.SERVER,observer_identification);
            http_common.header_line(str,http_common.CONTENTLEN,
                                    unb.to_unbounded_string(integer'image(html_err'length)));
            unb.append(str,ascii.cr & ascii.lf);
            unb.append(str,html_err);
        when others =>
            null;
        end case;
        return unb.to_string(str);
    end;

    function http(request: http_request.request) return string is
        use type ada.calendar.time;
        use type hl.reference;  -- for comparing references
        use type bl.reference;
        str,r,t: unb.unbounded_string;
        index: natural;
        bref:  bl.reference := bl.null_reference;
        href:  hl.reference := hl.null_reference;
        brk:   broker_queue;
        msg:   history_queue;
    begin
        if unb.to_string(request.uri.path) = "/" then
            str := unb.to_unbounded_string(build_response(ROOT));
            -- build response body
            r := unb.to_unbounded_string(html_inf);

            -- build the brokers list
            if not bl.is_empty(brokers) then
                bref := bl.last(brokers);
            end if;
            
            while bref /= bl.null_reference loop
                brk := bl.get_element(brokers,bref);

                -- generate html
                index := unb.index(r,"$BROKERS$");
                unb.replace_slice(r,index,index - 1 +
                                  unb.length(unb.to_unbounded_string("$BROKERS$")),
                                  "<li>Broker registrado como " &
                                  trim(unb.to_string(brk.id),ada.strings.both) &
                                  " con " & trim(natural'image(brk.money),ada.strings.both) &
                                  " euros y " & trim(natural'image(brk.stock),ada.strings.both) & 
                                  " acciones.</li>" & ascii.lf &
                                  "$BROKERS$");
                -- go next!
                bl.rewind(brokers,bref);
            end loop;

            -- delete last
            index := unb.index(r,"$BROKERS$");
            r := unb.delete(r,index,index - 1 +
                            unb.length(unb.to_unbounded_string("$BROKERS$")));

            -- build the history list
            if not hl.is_empty(history) then
                href := hl.last(history);
            end if;

            while href /= hl.null_reference loop
                -- get offer data
                msg := hl.get_element(history,href);
                case msg.t is
                when MOFFER =>
                    if msg.offer.kind = BUY then
                        t := unb.to_unbounded_string("compra");
                    else
                        t := unb.to_unbounded_string("venta");
                    end if;
    
                    -- generate html
                    index := unb.index(r,"$OPERS$");
                    unb.replace_slice(r,index,index - 1 +
                                      unb.length(unb.to_unbounded_string("$OPERS$")),
                                      "<li>Oferta: " &
                                      trim(unb.to_string(msg.offer.offerer_id),ada.strings.both) &
                                      ", " & utl.image(msg.tmstamp) & 
                                      ", " & trim(natural'image(msg.offer.stock_price),ada.strings.both) &
                                      " (precio/unidad), " &
                                      trim(natural'image(msg.offer.quantity),ada.strings.both) &
                                      ", " & unb.to_string(t) & ascii.lf & "</li>" & ascii.lf &
                                      "$OPERS$");
                when MACCEPT =>
                    -- generate html
                    index := unb.index(r,"$OPERS$");
                    unb.replace_slice(r,index,index - 1 +
                                      unb.length(unb.to_unbounded_string("$OPERS$")),
                                      "<li>Aceptaci&oacute;n de " &
                                      trim(unb.to_string(msg.accep.acceptor_id),ada.strings.both) &
                                      " a la oferta n&uacute;mero " & 
                                      trim(natural'image(msg.accep.offer_id),ada.strings.both) &
                                      " de " & trim(unb.to_string(msg.accep.offerer_id),ada.strings.both) &
                                      " a las " & utl.image(msg.tmstamp) & "</li>" & ascii.lf &
                                      "$OPERS$");
                when MGONG =>
                    -- generate html
                    index := unb.index(r,"$OPERS$");
                    unb.replace_slice(r,index,index - 1 +
                                      unb.length(unb.to_unbounded_string("$OPERS$")),
                                      "<li>Comienza la ronda " &
                                      trim(natural'image(msg.round),ada.strings.both) &
                                      " a las " & utl.image(msg.tmstamp) & "</li>" & ascii.lf &
                                      "$OPERS$");
                when MCOMMIT =>
                    -- generate html
                    index := unb.index(r,"$OPERS$");
                    unb.replace_slice(r,index,index - 1 +
                                      unb.length(unb.to_unbounded_string("$OPERS$")),
                                      "<li>Comprometida para " &
                                      trim(unb.to_string(msg.acceptor_id),ada.strings.both) &
                                      " la oferta " &
                                      trim(natural'image(msg.offer_id),ada.strings.both) &
                                      " realizada por " &
                                      trim(unb.to_string(msg.offerer_id),ada.strings.both) &
                                      " a las " & utl.image(msg.tmstamp) & "</li>" & ascii.lf &
                                      "$OPERS$");
                end case;

                -- go next!
                hl.rewind(history,href);
            end loop;

            -- delete last
            index := unb.index(r,"$OPERS$");
            r := unb.delete(r,index,index - 1 +
                            unb.length(unb.to_unbounded_string("$OPERS$")));

            http_common.header_line(str,http_common.CONTENTLEN,
                                    unb.to_unbounded_string(integer'image(
                                    unb.length(r))));
            unb.append(str,ascii.cr & ascii.lf);
            unb.append(str,r);

            return unb.to_string(str);
        else    -- any other page
            return build_response(ERROR);
        end if;

    exception
        when e: others =>
            utl.error("http: unexpected error: " & ada.exceptions.exception_message(e));
            return build_response(ERROR);
    end;

    procedure bind_multicast(addr: udp.end_point_type) is
    begin
        multicast_main_addr := addr;
        udp.bind(addr,hdl.main_handler'access);
    end;

end observer_control;